@@ -25,13 +25,15 @@ static void        soup_ntlm_nt_hash           (const char  *password,
 static char       *soup_ntlm_request           (void);
 static gboolean    soup_ntlm_parse_challenge   (const char  *challenge,
 						char       **nonce,
-						char       **default_domain);
+						char       **default_domain,
+						gboolean    *ntlmv2_session);
 static char       *soup_ntlm_response          (const char  *nonce, 
 						const char  *user,
 						guchar       nt_hash[21],
 						guchar       lm_hash[21],
 						const char  *host, 
-						const char  *domain);
+						const char  *domain,
+						gboolean     ntlmv2_session);
 
 typedef enum {
 	SOUP_NTLM_NEW,
@@ -46,6 +48,7 @@ typedef struct {
 	SoupNTLMState state;
 	char *nonce;
 	char *response_header;
+	gboolean ntlmv2_session;
 } SoupNTLMConnectionState;
 
 typedef enum {
@@ -73,7 +76,6 @@ typedef struct {
 	int fd_out;
 #endif
 } SoupAuthNTLMPrivate;
-#define SOUP_AUTH_NTLM_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_AUTH_NTLM, SoupAuthNTLMPrivate))
 
 #ifdef USE_NTLM_AUTH
 static gboolean ntlm_auth_available, ntlm_auth_debug;
@@ -91,13 +93,13 @@ static void sso_ntlm_close (SoupAuthNTLMPrivate *priv);
  * Since: 2.34
  */
 
-G_DEFINE_TYPE (SoupAuthNTLM, soup_auth_ntlm, SOUP_TYPE_CONNECTION_AUTH)
+G_DEFINE_TYPE_WITH_PRIVATE (SoupAuthNTLM, soup_auth_ntlm, SOUP_TYPE_CONNECTION_AUTH)
 
 static void
 soup_auth_ntlm_init (SoupAuthNTLM *ntlm)
 {
 #ifdef USE_NTLM_AUTH
-	SoupAuthNTLMPrivate *priv = SOUP_AUTH_NTLM_GET_PRIVATE (ntlm);
+	SoupAuthNTLMPrivate *priv = soup_auth_ntlm_get_instance_private (ntlm);
 	const char *username = NULL, *slash;
 
 	priv->sso_available = TRUE;
@@ -122,7 +124,7 @@ soup_auth_ntlm_init (SoupAuthNTLM *ntlm)
 static void
 soup_auth_ntlm_finalize (GObject *object)
 {
-	SoupAuthNTLMPrivate *priv = SOUP_AUTH_NTLM_GET_PRIVATE (object);
+	SoupAuthNTLMPrivate *priv = soup_auth_ntlm_get_instance_private (SOUP_AUTH_NTLM (object));
 
 	g_free (priv->username);
 	g_free (priv->domain);
@@ -285,7 +287,8 @@ static gboolean
 soup_auth_ntlm_update_connection (SoupConnectionAuth *auth, SoupMessage *msg,
 				  const char *auth_header, gpointer state)
 {
-	SoupAuthNTLMPrivate *priv = SOUP_AUTH_NTLM_GET_PRIVATE (auth);
+	SoupAuthNTLM *auth_ntlm = SOUP_AUTH_NTLM (auth);
+	SoupAuthNTLMPrivate *priv = soup_auth_ntlm_get_instance_private (auth_ntlm);
 	SoupNTLMConnectionState *conn = state;
 	gboolean success = TRUE;
 
@@ -300,12 +303,15 @@ soup_auth_ntlm_update_connection (SoupConnectionAuth *auth, SoupMessage *msg,
 	if (conn->state > SOUP_NTLM_SENT_REQUEST) {
 		if (priv->password_state == SOUP_NTLM_PASSWORD_ACCEPTED) {
 			/* We know our password is correct, so a 401
-			 * means "permission denied". Since the conn
-			 * state is now FAILED, the auth is no longer
-			 * is_ready() for this message, so this will
-			 * cause a "retrying" authenticate signal.
+			 * means "permission denied". The code can't deal
+			 * with re-authenticating correctly, so make sure
+			 * we don't try.
 			 */
 			conn->state = SOUP_NTLM_FAILED;
+			if (soup_message_is_keepalive (msg)) {
+				soup_message_headers_append (msg->response_headers,
+							     "Connection", "close");
+			}
 			return TRUE;
 		}
 
@@ -326,8 +332,14 @@ soup_auth_ntlm_update_connection (SoupConnectionAuth *auth, SoupMessage *msg,
 	if (conn->state == SOUP_NTLM_NEW && !auth_header[4])
 		return TRUE;
 
+	if (!auth_header[4] || !auth_header[5]) {
+		conn->state = SOUP_NTLM_FAILED;
+		return FALSE;
+	}
+
 	if (!soup_ntlm_parse_challenge (auth_header + 5, &conn->nonce,
-					priv->domain ? NULL : &priv->domain)) {
+					priv->domain ? NULL : &priv->domain,
+					&conn->ntlmv2_session)) {
 		conn->state = SOUP_NTLM_FAILED;
 		return FALSE;
 	}
@@ -352,6 +364,7 @@ soup_auth_ntlm_update_connection (SoupConnectionAuth *auth, SoupMessage *msg,
 			conn->state = SOUP_NTLM_SSO_FAILED;
 			success = FALSE;
 		} else if (!g_ascii_strcasecmp (response, "PW")) {
+			conn->state = SOUP_NTLM_SSO_FAILED;
 			priv->sso_available = FALSE;
 			g_free (response);
 		} else {
@@ -392,7 +405,8 @@ static void
 soup_auth_ntlm_authenticate (SoupAuth *auth, const char *username,
 			     const char *password)
 {
-	SoupAuthNTLMPrivate *priv = SOUP_AUTH_NTLM_GET_PRIVATE (auth);
+	SoupAuthNTLM *auth_ntlm = SOUP_AUTH_NTLM (auth);
+	SoupAuthNTLMPrivate *priv = soup_auth_ntlm_get_instance_private (auth_ntlm);
 	const char *slash;
 
 	g_return_if_fail (username != NULL);
@@ -421,7 +435,8 @@ soup_auth_ntlm_authenticate (SoupAuth *auth, const char *username,
 static gboolean
 soup_auth_ntlm_is_authenticated (SoupAuth *auth)
 {
-	SoupAuthNTLMPrivate *priv = SOUP_AUTH_NTLM_GET_PRIVATE (auth);
+	SoupAuthNTLM *auth_ntlm = SOUP_AUTH_NTLM (auth);
+	SoupAuthNTLMPrivate *priv = soup_auth_ntlm_get_instance_private (auth_ntlm);
 
 	return (priv->password_state != SOUP_NTLM_PASSWORD_NONE &&
 		priv->password_state != SOUP_NTLM_PASSWORD_REJECTED);
@@ -432,7 +447,8 @@ soup_auth_ntlm_is_connection_ready (SoupConnectionAuth *auth,
 				    SoupMessage        *msg,
 				    gpointer            state)
 {
-	SoupAuthNTLMPrivate *priv = SOUP_AUTH_NTLM_GET_PRIVATE (auth);
+	SoupAuthNTLM *auth_ntlm = SOUP_AUTH_NTLM (auth);
+	SoupAuthNTLMPrivate *priv = soup_auth_ntlm_get_instance_private (auth_ntlm);
 	SoupNTLMConnectionState *conn = state;
 
 	if (priv->password_state == SOUP_NTLM_PASSWORD_REJECTED)
@@ -448,6 +464,7 @@ static void
 got_final_auth_result (SoupMessage *msg, gpointer data)
 {
 	SoupAuth *auth = data;
+	SoupAuthNTLMPrivate *priv = soup_auth_ntlm_get_instance_private (SOUP_AUTH_NTLM (auth));
 
 	g_signal_handlers_disconnect_by_func (msg, G_CALLBACK (got_final_auth_result), auth);
 
@@ -455,7 +472,7 @@ got_final_auth_result (SoupMessage *msg, gpointer data)
 		return;
 
 	if (msg->status_code != SOUP_STATUS_UNAUTHORIZED)
-		SOUP_AUTH_NTLM_GET_PRIVATE (auth)->password_state = SOUP_NTLM_PASSWORD_ACCEPTED;
+		priv->password_state = SOUP_NTLM_PASSWORD_ACCEPTED;
 }
 
 static char *
@@ -463,7 +480,8 @@ soup_auth_ntlm_get_connection_authorization (SoupConnectionAuth *auth,
 					     SoupMessage        *msg,
 					     gpointer            state)
 {
-	SoupAuthNTLMPrivate *priv = SOUP_AUTH_NTLM_GET_PRIVATE (auth);
+	SoupAuthNTLM *auth_ntlm = SOUP_AUTH_NTLM (auth);
+	SoupAuthNTLMPrivate *priv = soup_auth_ntlm_get_instance_private (auth_ntlm);
 	SoupNTLMConnectionState *conn = state;
 	char *header = NULL;
 
@@ -482,7 +500,7 @@ soup_auth_ntlm_get_connection_authorization (SoupConnectionAuth *auth,
 					priv->sso_available = FALSE;
 				}
 			} else {
-				g_warning ("NTLM single-sign-on using %s failed", NTLM_AUTH);
+				g_debug ("NTLM single-sign-on using %s failed", NTLM_AUTH);
 			}
 		}
 		/* If NTLM single-sign-on fails, go back to original
@@ -502,7 +520,8 @@ soup_auth_ntlm_get_connection_authorization (SoupConnectionAuth *auth,
 						     priv->nt_hash,
 						     priv->lm_hash,
 						     NULL,
-						     priv->domain);
+						     priv->domain,
+						     conn->ntlmv2_session);
 		}
 		g_clear_pointer (&conn->nonce, g_free);
 		conn->state = SOUP_NTLM_SENT_RESPONSE;
@@ -517,7 +536,7 @@ soup_auth_ntlm_get_connection_authorization (SoupConnectionAuth *auth,
 #ifdef USE_NTLM_AUTH
 	case SOUP_NTLM_SSO_FAILED:
 		/* Restart request without SSO */
-		g_warning ("NTLM single-sign-on by using %s failed", NTLM_AUTH);
+		g_debug ("NTLM single-sign-on by using %s failed", NTLM_AUTH);
 		priv->sso_available = FALSE;
 		header = soup_ntlm_request ();
 		conn->state = SOUP_NTLM_SENT_REQUEST;
@@ -537,8 +556,6 @@ soup_auth_ntlm_class_init (SoupAuthNTLMClass *auth_ntlm_class)
 	SoupConnectionAuthClass *connauth_class = SOUP_CONNECTION_AUTH_CLASS (auth_ntlm_class);
 	GObjectClass *object_class = G_OBJECT_CLASS (auth_ntlm_class);
 
-	g_type_class_add_private (auth_ntlm_class, sizeof (SoupAuthNTLMPrivate));
-
 	auth_class->scheme_name = "NTLM";
 	auth_class->strength = 3;
 
@@ -631,8 +648,12 @@ typedef struct {
 #define NTLM_CHALLENGE_NONCE_LENGTH          8
 #define NTLM_CHALLENGE_DOMAIN_STRING_OFFSET 12
 
+#define NTLM_CHALLENGE_FLAGS_OFFSET         20
+#define NTLM_FLAGS_NEGOTIATE_NTLMV2 0x00080000
+
 #define NTLM_RESPONSE_HEADER "NTLMSSP\x00\x03\x00\x00\x00"
 #define NTLM_RESPONSE_FLAGS 0x8201
+#define NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY 0x00080000
 
 typedef struct {
         guchar     header[12];
@@ -658,17 +679,19 @@ ntlm_set_string (NTLMString *string, int *offset, int len)
 static char *
 soup_ntlm_request (void)
 {
-	return g_strdup ("NTLM TlRMTVNTUAABAAAABYIAAAAAAAAAAAAAAAAAAAAAAAAAAAAAMAAAAAAAAAAwAAAA");
+	return g_strdup ("NTLM TlRMTVNTUAABAAAABYIIAAAAAAAAAAAAAAAAAAAAAAAAAAAAMAAAAAAAAAAwAAAA");
 }
 
 static gboolean
 soup_ntlm_parse_challenge (const char *challenge,
 			   char      **nonce,
-			   char      **default_domain)
+			   char      **default_domain,
+			   gboolean   *ntlmv2_session)
 {
 	gsize clen;
 	NTLMString domain;
 	guchar *chall;
+	guint32 flags;
 
 	chall = g_base64_decode (challenge, &clen);
 	if (clen < NTLM_CHALLENGE_DOMAIN_STRING_OFFSET ||
@@ -677,6 +700,10 @@ soup_ntlm_parse_challenge (const char *challenge,
 		return FALSE;
 	}
 
+	memcpy (&flags, chall + NTLM_CHALLENGE_FLAGS_OFFSET, sizeof(flags));
+	flags = GUINT_FROM_LE (flags);
+	*ntlmv2_session = (flags & NTLM_FLAGS_NEGOTIATE_NTLMV2) ? TRUE : FALSE;
+
 	if (default_domain) {
 		memcpy (&domain, chall + NTLM_CHALLENGE_DOMAIN_STRING_OFFSET, sizeof (domain));
 		domain.length = GUINT16_FROM_LE (domain.length);
@@ -701,13 +728,48 @@ soup_ntlm_parse_challenge (const char *challenge,
 	return TRUE;
 }
 
+static void
+calc_ntlm2_session_response (const char *nonce,
+			     guchar      nt_hash[21],
+			     guchar      lm_hash[21],
+			     guchar     *lm_resp,
+			     gsize       lm_resp_sz,
+			     guchar     *nt_resp)
+{
+	guint32 client_nonce[2];
+	guchar ntlmv2_hash[16];
+	GChecksum *ntlmv2_cksum;
+	gsize ntlmv2_hash_sz = sizeof (ntlmv2_hash);
+
+	/* FIXME: if GLib ever gets a more secure random number
+	 * generator, use it here
+	 */
+	client_nonce[0] = g_random_int();
+	client_nonce[1] = g_random_int();
+
+	ntlmv2_cksum = g_checksum_new (G_CHECKSUM_MD5);
+	g_checksum_update (ntlmv2_cksum, (const guchar *) nonce, 8);
+	g_checksum_update (ntlmv2_cksum, (const guchar *) client_nonce, sizeof (client_nonce));
+	g_checksum_get_digest (ntlmv2_cksum, ntlmv2_hash, &ntlmv2_hash_sz);
+	g_checksum_free (ntlmv2_cksum);
+
+	/* Send the padded client nonce as a fake lm_resp */
+	memset (lm_resp, 0, lm_resp_sz);
+	memcpy (lm_resp, client_nonce, sizeof (client_nonce));
+
+	/* Compute nt_hash as usual but with a new nonce */
+	calc_response (nt_hash, ntlmv2_hash, nt_resp);
+}
+
+
 static char *
 soup_ntlm_response (const char *nonce, 
 		    const char *user,
 		    guchar      nt_hash[21],
 		    guchar      lm_hash[21],
 		    const char *host, 
-		    const char *domain)
+		    const char *domain,
+		    gboolean    ntlmv2_session)
 {
 	int offset;
 	gsize hlen, dlen, ulen;
@@ -717,12 +779,20 @@ soup_ntlm_response (const char *nonce,
 	char *out, *p;
 	int state, save;
 
-	calc_response (nt_hash, (guchar *)nonce, nt_resp);
-	calc_response (lm_hash, (guchar *)nonce, lm_resp);
+	if (ntlmv2_session) {
+		calc_ntlm2_session_response (nonce, nt_hash, lm_hash,
+					     lm_resp, sizeof(lm_resp), nt_resp);
+	} else {
+		/* Compute a regular response */
+		calc_response (nt_hash, (guchar *) nonce, nt_resp);
+		calc_response (lm_hash, (guchar *) nonce, lm_resp);
+	}
 
 	memset (&resp, 0, sizeof (resp));
 	memcpy (resp.header, NTLM_RESPONSE_HEADER, sizeof (resp.header));
 	resp.flags = GUINT32_TO_LE (NTLM_RESPONSE_FLAGS);
+	if (ntlmv2_session)
+		resp.flags |= GUINT32_TO_LE (NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY);
 
 	offset = sizeof (resp);
 
@@ -740,7 +810,7 @@ soup_ntlm_response (const char *nonce,
 	ntlm_set_string (&resp.nt_resp, &offset, sizeof (nt_resp));
 
 	out = g_malloc (((offset + 3) * 4) / 3 + 6);
-	strncpy (out, "NTLM ", 5);
+	memcpy (out, "NTLM ", 5);
 	p = out + 5;
 
 	state = save = 0;
@@ -835,7 +905,8 @@ md4sum (const unsigned char *in, int nbytes, unsigned char digest[16])
 	guint32 A, B, C, D, AA, BB, CC, DD, X[16];
 	int pbytes, nbits = nbytes * 8, i, j;
 
-	pbytes = (120 - (nbytes % 64)) % 64;
+	/* There is *always* padding of at least one bit. */
+	pbytes = ((119 - (nbytes % 64)) % 64) + 1;
 	M = alloca (nbytes + pbytes + 8);
 	memcpy (M, in, nbytes);
 	memset (M + nbytes, 0, pbytes + 8);
