@@ -11,7 +11,6 @@
 
 #include "soup-client-input-stream.h"
 #include "soup.h"
-#include "soup-marshal.h"
 #include "soup-message-private.h"
 
 struct _SoupClientInputStreamPrivate {
@@ -19,7 +18,7 @@ struct _SoupClientInputStreamPrivate {
 };
 
 enum {
-	EOF,
+	SIGNAL_EOF,
 	LAST_SIGNAL
 };
 
@@ -35,15 +34,14 @@ static GPollableInputStreamInterface *soup_client_input_stream_parent_pollable_i
 static void soup_client_input_stream_pollable_init (GPollableInputStreamInterface *pollable_interface, gpointer interface_data);
 
 G_DEFINE_TYPE_WITH_CODE (SoupClientInputStream, soup_client_input_stream, SOUP_TYPE_FILTER_INPUT_STREAM,
+                         G_ADD_PRIVATE (SoupClientInputStream)
 			 G_IMPLEMENT_INTERFACE (G_TYPE_POLLABLE_INPUT_STREAM,
 						soup_client_input_stream_pollable_init))
 
 static void
 soup_client_input_stream_init (SoupClientInputStream *stream)
 {
-	stream->priv = G_TYPE_INSTANCE_GET_PRIVATE (stream,
-						    SOUP_TYPE_CLIENT_INPUT_STREAM,
-						    SoupClientInputStreamPrivate);
+	stream->priv = soup_client_input_stream_get_instance_private (stream);
 }
 
 static void
@@ -101,7 +99,7 @@ soup_client_input_stream_read_fn (GInputStream  *stream,
 		read_fn (stream, buffer, count, cancellable, error);
 
 	if (nread == 0)
-		g_signal_emit (stream, signals[EOF], 0);
+		g_signal_emit (stream, signals[SIGNAL_EOF], 0);
 
 	return nread;
 }
@@ -118,7 +116,7 @@ soup_client_input_stream_read_nonblocking (GPollableInputStream  *stream,
 		read_nonblocking (stream, buffer, count, error);
 
 	if (nread == 0)
-		g_signal_emit (stream, signals[EOF], 0);
+		g_signal_emit (stream, signals[SIGNAL_EOF], 0);
 
 	return nread;
 }
@@ -129,9 +127,12 @@ soup_client_input_stream_close_fn (GInputStream  *stream,
 				   GError       **error)
 {
 	SoupClientInputStream *cistream = SOUP_CLIENT_INPUT_STREAM (stream);
+	gboolean success;
 
-	return soup_message_io_run_until_finish (cistream->priv->msg,
-						 cancellable, error);
+	success = soup_message_io_run_until_finish (cistream->priv->msg, TRUE,
+						    NULL, error);
+	soup_message_io_finished (cistream->priv->msg);
+	return success;
 }
 
 static gboolean
@@ -151,7 +152,7 @@ close_async_ready (SoupMessage *msg, gpointer user_data)
 	SoupClientInputStream *cistream = g_task_get_source_object (task);
 	GError *error = NULL;
 
-	if (!soup_message_io_run_until_finish (cistream->priv->msg,
+	if (!soup_message_io_run_until_finish (cistream->priv->msg, FALSE,
 					       g_task_get_cancellable (task),
 					       &error) &&
 	    g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
@@ -159,6 +160,8 @@ close_async_ready (SoupMessage *msg, gpointer user_data)
 		return TRUE;
 	}
 
+	soup_message_io_finished (cistream->priv->msg);
+
 	if (error) {
 		g_task_return_error (task, error);
 		g_object_unref (task);
@@ -211,8 +214,6 @@ soup_client_input_stream_class_init (SoupClientInputStreamClass *stream_class)
 	GObjectClass *object_class = G_OBJECT_CLASS (stream_class);
 	GInputStreamClass *input_stream_class = G_INPUT_STREAM_CLASS (stream_class);
 
-	g_type_class_add_private (stream_class, sizeof (SoupClientInputStreamPrivate));
-
 	object_class->finalize = soup_client_input_stream_finalize;
 	object_class->set_property = soup_client_input_stream_set_property;
 	object_class->get_property = soup_client_input_stream_get_property;
@@ -222,13 +223,13 @@ soup_client_input_stream_class_init (SoupClientInputStreamClass *stream_class)
 	input_stream_class->close_async = soup_client_input_stream_close_async;
 	input_stream_class->close_finish = soup_client_input_stream_close_finish;
 
-	signals[EOF] =
+	signals[SIGNAL_EOF] =
 		g_signal_new ("eof",
 			      G_OBJECT_CLASS_TYPE (object_class),
 			      G_SIGNAL_RUN_LAST,
 			      0,
 			      NULL, NULL,
-			      _soup_marshal_NONE__NONE,
+			      NULL,
 			      G_TYPE_NONE, 0);
 
 	g_object_class_install_property (
