@@ -44,9 +44,9 @@
  * > Host: localhost
  * > Content-Type: text/plain
  * > Connection: close
- * > 
+ * >
  * > This is a test.
- *   
+ *
  * &lt; HTTP/1.1 201 Created
  * &lt; Soup-Debug-Timestamp: 1200171744
  * &lt; Soup-Debug: SoupMessage 1 (0x617000)
@@ -88,13 +88,6 @@
  * event of the #SoupMessage::finished signal.
  **/
 
-static SoupSessionFeatureInterface *soup_logger_default_feature_interface;
-static void soup_logger_session_feature_init (SoupSessionFeatureInterface *feature_interface, gpointer interface_data);
-
-G_DEFINE_TYPE_WITH_CODE (SoupLogger, soup_logger, G_TYPE_OBJECT,
-			 G_IMPLEMENT_INTERFACE (SOUP_TYPE_SESSION_FEATURE,
-						soup_logger_session_feature_init))
-
 typedef struct {
 	/* We use a mutex so that if requests are being run in
 	 * multiple threads, we don't mix up the output.
@@ -120,12 +113,28 @@ typedef struct {
 	gpointer            printer_data;
 	GDestroyNotify      printer_dnotify;
 } SoupLoggerPrivate;
-#define SOUP_LOGGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_LOGGER, SoupLoggerPrivate))
+
+enum {
+	PROP_0,
+
+	PROP_LEVEL,
+	PROP_MAX_BODY_SIZE,
+
+	LAST_PROP
+};
+
+static SoupSessionFeatureInterface *soup_logger_default_feature_interface;
+static void soup_logger_session_feature_init (SoupSessionFeatureInterface *feature_interface, gpointer interface_data);
+
+G_DEFINE_TYPE_WITH_CODE (SoupLogger, soup_logger, G_TYPE_OBJECT,
+                         G_ADD_PRIVATE (SoupLogger)
+			 G_IMPLEMENT_INTERFACE (SOUP_TYPE_SESSION_FEATURE,
+						soup_logger_session_feature_init))
 
 static void
 soup_logger_init (SoupLogger *logger)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	g_mutex_init (&priv->lock);
 	priv->tag = g_quark_from_static_string (g_strdup_printf ("SoupLogger-%p", logger));
@@ -135,7 +144,8 @@ soup_logger_init (SoupLogger *logger)
 static void
 soup_logger_finalize (GObject *object)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (object);
+	SoupLogger *logger = SOUP_LOGGER (object);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	g_hash_table_destroy (priv->ids);
 
@@ -151,14 +161,104 @@ soup_logger_finalize (GObject *object)
 	G_OBJECT_CLASS (soup_logger_parent_class)->finalize (object);
 }
 
+static void
+soup_logger_set_property (GObject *object, guint prop_id,
+			  const GValue *value, GParamSpec *pspec)
+{
+	SoupLogger *logger = SOUP_LOGGER (object);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
+
+	switch (prop_id) {
+	case PROP_LEVEL:
+		priv->level = g_value_get_enum (value);
+		break;
+	case PROP_MAX_BODY_SIZE:
+		priv->max_body_size = g_value_get_int (value);
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+		break;
+	}
+}
+
+static void
+soup_logger_get_property (GObject *object, guint prop_id,
+			   GValue *value, GParamSpec *pspec)
+{
+	SoupLogger *logger = SOUP_LOGGER (object);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
+
+	switch (prop_id) {
+	case PROP_LEVEL:
+		g_value_set_enum (value, priv->level);
+		break;
+	case PROP_MAX_BODY_SIZE:
+		g_value_set_int (value, priv->max_body_size);
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+		break;
+	}
+}
+
 static void
 soup_logger_class_init (SoupLoggerClass *logger_class)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (logger_class);
 
-	g_type_class_add_private (logger_class, sizeof (SoupLoggerPrivate));
-
 	object_class->finalize = soup_logger_finalize;
+	object_class->set_property = soup_logger_set_property;
+	object_class->get_property = soup_logger_get_property;
+
+	/* properties */
+	/**
+	 * SoupLogger:level:
+	 *
+	 * The level of logging output
+	 *
+	 * Since: 2.56
+	 */
+	/**
+	 * SOUP_LOGGER_LEVEL:
+	 *
+	 * Alias for the #SoupLogger:level property, qv.
+	 *
+	 * Since: 2.56
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_LEVEL,
+		g_param_spec_enum (SOUP_LOGGER_LEVEL,
+				    "Level",
+				    "The level of logging output",
+				    SOUP_TYPE_LOGGER_LOG_LEVEL,
+				    SOUP_LOGGER_LOG_MINIMAL,
+				    G_PARAM_READWRITE));
+
+	/**
+	 * SoupLogger:max-body-size:
+	 *
+	 * If #SoupLogger:level is %SOUP_LOGGER_LOG_BODY, this gives
+	 * the maximum number of bytes of the body that will be logged.
+	 * (-1 means "no limit".)
+	 *
+	 * Since: 2.56
+	 */
+	/**
+	 * SOUP_LOGGER_MAX_BODY_SIZE:
+	 *
+	 * Alias for the #SoupLogger:max-body-size property, qv.
+	 *
+	 * Since: 2.56
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_MAX_BODY_SIZE,
+		g_param_spec_int (SOUP_LOGGER_MAX_BODY_SIZE,
+				    "Max Body Size",
+				    "The maximum body size to output",
+				    -1,
+				    G_MAXINT,
+				    -1,
+				    G_PARAM_READWRITE));
 }
 
 /**
@@ -189,18 +289,12 @@ soup_logger_class_init (SoupLoggerClass *logger_class)
  * Returns: a new #SoupLogger
  **/
 SoupLogger *
-soup_logger_new (SoupLoggerLogLevel level, int max_body_size) 
+soup_logger_new (SoupLoggerLogLevel level, int max_body_size)
 {
-	SoupLogger *logger;
-	SoupLoggerPrivate *priv;
-
-	logger = g_object_new (SOUP_TYPE_LOGGER, NULL);
-
-	priv = SOUP_LOGGER_GET_PRIVATE (logger);
-	priv->level = level;
-	priv->max_body_size = max_body_size;
-
-	return logger;
+	return g_object_new (SOUP_TYPE_LOGGER,
+			     SOUP_LOGGER_LEVEL, level,
+			     SOUP_LOGGER_MAX_BODY_SIZE, max_body_size,
+			     NULL);
 }
 
 /**
@@ -239,7 +333,7 @@ soup_logger_set_request_filter (SoupLogger       *logger,
 				gpointer          filter_data,
 				GDestroyNotify    destroy)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	priv->request_filter         = request_filter;
 	priv->request_filter_data    = filter_data;
@@ -265,7 +359,7 @@ soup_logger_set_response_filter (SoupLogger       *logger,
 				 gpointer          filter_data,
 				 GDestroyNotify    destroy)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	priv->response_filter         = response_filter;
 	priv->response_filter_data    = filter_data;
@@ -311,7 +405,7 @@ soup_logger_set_printer (SoupLogger        *logger,
 			 gpointer           printer_data,
 			 GDestroyNotify     destroy)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	priv->printer         = printer;
 	priv->printer_data    = printer_data;
@@ -321,7 +415,7 @@ soup_logger_set_printer (SoupLogger        *logger,
 static guint
 soup_logger_get_id (SoupLogger *logger, gpointer object)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	return GPOINTER_TO_UINT (g_object_get_qdata (object, priv->tag));
 }
@@ -329,7 +423,7 @@ soup_logger_get_id (SoupLogger *logger, gpointer object)
 static guint
 soup_logger_set_id (SoupLogger *logger, gpointer object)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 	gpointer klass = G_OBJECT_GET_CLASS (object);
 	gpointer id;
 
@@ -385,7 +479,7 @@ static void
 soup_logger_print (SoupLogger *logger, SoupLoggerLogLevel level,
 		   char direction, const char *format, ...)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 	va_list args;
 	char *data, *line, *end;
 
@@ -449,7 +543,7 @@ static void
 print_request (SoupLogger *logger, SoupMessage *msg,
 	       SoupSocket *socket, gboolean restarted)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 	SoupLoggerLogLevel log_level;
 	SoupMessageHeadersIter iter;
 	const char *name, *value;
@@ -552,7 +646,7 @@ print_request (SoupLogger *logger, SoupMessage *msg,
 static void
 print_response (SoupLogger *logger, SoupMessage *msg)
 {
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 	SoupLoggerLogLevel log_level;
 	SoupMessageHeadersIter iter;
 	const char *name, *value;
@@ -600,12 +694,12 @@ static void
 finished (SoupMessage *msg, gpointer user_data)
 {
 	SoupLogger *logger = user_data;
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	g_mutex_lock (&priv->lock);
 
 	print_response (logger, msg);
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "\n");
 
 	g_mutex_unlock (&priv->lock);
 }
@@ -614,13 +708,13 @@ static void
 got_informational (SoupMessage *msg, gpointer user_data)
 {
 	SoupLogger *logger = user_data;
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	g_mutex_lock (&priv->lock);
 
 	g_signal_handlers_disconnect_by_func (msg, finished, logger);
 	print_response (logger, msg);
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "\n");
 
 	if (msg->status_code == SOUP_STATUS_CONTINUE && msg->request_body->data) {
 		SoupLoggerLogLevel log_level;
@@ -639,7 +733,7 @@ got_informational (SoupMessage *msg, gpointer user_data)
 					   "%s", msg->request_body->data);
 		}
 
-		soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
+		soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "\n");
 	}
 
 	g_mutex_unlock (&priv->lock);
@@ -649,13 +743,13 @@ static void
 got_body (SoupMessage *msg, gpointer user_data)
 {
 	SoupLogger *logger = user_data;
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 
 	g_mutex_lock (&priv->lock);
 
 	g_signal_handlers_disconnect_by_func (msg, finished, logger);
 	print_response (logger, msg);
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "\n");
 
 	g_mutex_unlock (&priv->lock);
 }
@@ -664,7 +758,7 @@ static void
 starting (SoupMessage *msg, gpointer user_data)
 {
 	SoupLogger *logger = SOUP_LOGGER (user_data);
-	SoupLoggerPrivate *priv = SOUP_LOGGER_GET_PRIVATE (logger);
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (logger);
 	gboolean restarted;
 	guint msg_id;
 	SoupConnection *conn;
@@ -687,7 +781,7 @@ starting (SoupMessage *msg, gpointer user_data)
 		soup_logger_set_id (logger, socket);
 
 	print_request (logger, msg, socket, restarted);
-	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "");
+	soup_logger_print (logger, SOUP_LOGGER_LOG_MINIMAL, ' ', "\n");
 }
 
 static void
@@ -728,7 +822,9 @@ static void
 soup_logger_feature_attach (SoupSessionFeature *feature,
 			    SoupSession *session)
 {
-	SOUP_LOGGER_GET_PRIVATE (feature)->session = session;
+	SoupLoggerPrivate *priv = soup_logger_get_instance_private (SOUP_LOGGER (feature));
+
+	priv->session = session;
 
 	soup_logger_default_feature_interface->attach (feature, session);
 }
