@@ -4,14 +4,14 @@
  * Copyright 2011 Red Hat, Inc.
  */
 
-#ifndef SOUP_MISC_PRIVATE_H
-#define SOUP_MISC_PRIVATE_H 1
+#ifndef __SOUP_MISC_PRIVATE_H__
+#define __SOUP_MISC_PRIVATE_H__ 1
 
 #include "soup-message-headers.h"
 
 char *soup_uri_decoded_copy (const char *str, int length, int *decoded_length);
 char *soup_uri_to_string_internal (SoupURI *uri, gboolean just_path_and_query,
-				   gboolean force_port);
+				   gboolean include_password, gboolean force_port);
 gboolean soup_uri_is_http (SoupURI *uri, char **aliases);
 gboolean soup_uri_is_https (SoupURI *uri, char **aliases);
 
@@ -30,9 +30,10 @@ gboolean soup_uri_is_https (SoupURI *uri, char **aliases);
 					   method == SOUP_METHOD_PUT || \
 					   method == SOUP_METHOD_DELETE)
 
-GSource *soup_add_completion_reffed (GMainContext *async_context,
-				     GSourceFunc   function,
-				     gpointer      data);
+GSource *soup_add_completion_reffed (GMainContext   *async_context,
+				     GSourceFunc     function,
+				     gpointer        data,
+				     GDestroyNotify  dnotify);
 
 guint soup_message_headers_get_ranges_internal (SoupMessageHeaders  *hdrs,
 						goffset              total_length,
@@ -42,4 +43,7 @@ guint soup_message_headers_get_ranges_internal (SoupMessageHeaders  *hdrs,
 
 SoupAddress *soup_address_new_from_gsockaddr (GSocketAddress *addr);
 
-#endif /* SOUP_MISC_PRIVATE_H */
+gboolean           soup_host_matches_host    (const gchar *host,
+					      const gchar *compare_with);
+
+#endif /* __SOUP_MISC_PRIVATE_H__ */
